# passerelle - uniform access to multiple data sources and services
# Copyright (C) 2019 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
import mock
import os

import pytest

from passerelle.contrib.toulouse_axel.models import (
    AxelError,
    Link,
    ToulouseAxel,
    ref_famille_dui,
    ref_verif_dui,
)
import utils


@pytest.fixture
def resource(db):
    return utils.make_resource(
        ToulouseAxel,
        slug='test',
        wsdl_url='http://example.net/AXEL_WS/AxelWS.php?wsdl')


@pytest.fixture
def params():
    return {
        'IDDUI': 'XXX',
        'PRENOM': 'John',
        'NOM': 'Doe',
        'NAISSANCE': '2010-10-10',
    }


def test_operation_status_error(resource):
    resp = '''
    <?xml version="1.0"?>
    <PORTAILSERVICE>
      <RESULTAT>
        <TYPE>RefVerifDui</TYPE>
        <STATUS>NOK</STATUS>
        <COMMENTAIRES><![CDATA[Foo reason]]></COMMENTAIRES>
      </RESULTAT>
      <DATA>
        <PORTAIL/>
      </DATA>
    </PORTAILSERVICE>
    '''.strip()
    with mock.patch('passerelle.contrib.toulouse_axel.models.ToulouseAxel.soap_client') as client:
        client.return_value.service.getData.return_value = resp
        with pytest.raises(AxelError, match='Foo reason'):
            ref_verif_dui(resource, {
                'PORTAIL': {
                    'DUI': {
                        'PRENOM': 'John',
                        'NOM': 'Doe',
                        'NAISSANCE': '2010-10-10',
                    }
                }
            })


@contextmanager
def mock_getdata(content, operation):
    with mock.patch('passerelle.contrib.toulouse_axel.models.ToulouseAxel.soap_client') as client:
        resp = '''
        <?xml version="1.0"?>
        <PORTAILSERVICE>
          <RESULTAT>
            <TYPE>%s</TYPE>
            <STATUS>OK</STATUS>
            <DATE>10/10/2010 10:10:01</DATE>
            <COMMENTAIRES><![CDATA[]]></COMMENTAIRES>
          </RESULTAT>
          <DATA>
            %s
          </DATA>
        </PORTAILSERVICE>
        '''.strip() % (operation, content)
        client.return_value.service.getData.return_value = resp
        yield


@pytest.mark.parametrize('content', [
    '<PORTAIL><DUI/></PORTAIL>',
    '<PORTAIL><DUI><CODE>foo</CODE></DUI></PORTAIL>',
    '<PORTAIL><DUI><CODE>42</CODE></DUI></PORTAIL>',
])
def test_operation_ref_verif_dui(resource, content):
    with mock_getdata(content, 'RefVerifDui'):
        with pytest.raises(AxelError):
            ref_verif_dui(resource, {
                'PORTAIL': {
                    'DUI': {
                        'PRENOM': 'John',
                        'NOM': 'Doe',
                        'NAISSANCE': '2010-10-10',
                    }
                }
            })


@pytest.mark.parametrize('content', [
    '<PORTAIL><DUI/></PORTAIL>',
])
def test_operation_ref_famille_dui(resource, content):
    with mock_getdata(content, 'RefFamilleDui'):
        with pytest.raises(AxelError):
            ref_famille_dui(resource, {
                'PORTAIL': {
                    'DUI': {
                        'IDDUI': 'XXX',
                    }
                }
            })


def test_link_endpoint_axel_error(app, resource, params):
    with mock.patch('passerelle.contrib.toulouse_axel.models.ref_verif_dui') as operation:
        operation.side_effect = AxelError('FooBar')
        resp = app.post_json('/toulouse-axel/test/link?NameID=yyy', params=params)
        assert resp.json['err_desc'] == "Axel error: FooBar"


@pytest.mark.parametrize('xml_response', [
    '<IDDUI>XXX</IDDUI><IDPERSONNE/><CODE>0</CODE>',
    '<IDDUI>XXX</IDDUI><IDPERSONNE>42</IDPERSONNE><CODE>1</CODE>',
    '<IDDUI>XXX</IDDUI><IDPERSONNE>42</IDPERSONNE><CODE>4</CODE>',
])
def test_link_endpoint_no_result(app, resource, params, xml_response):
    content = '''<PORTAIL>
    <DUI>
        %s
    </DUI>
</PORTAIL>''' % xml_response
    with mock_getdata(content, 'RefVerifDui'):
        resp = app.post_json('/toulouse-axel/test/link?NameID=yyy', params=params)
    assert resp.json['err_desc'] == "Person not found"


def test_link_endpoint_conflict(app, resource, params):
    content = '''<PORTAIL>
    <DUI>
        <IDDUI>XXX</IDDUI>
        <IDPERSONNE>42</IDPERSONNE>
        <CODE>2</CODE>
    </DUI>
</PORTAIL>'''
    # existing link but dui is wrong
    link = Link.objects.create(resource=resource, name_id='yyy', dui='YYY', person_id='42')
    with mock_getdata(content, 'RefVerifDui'):
        resp = app.post_json('/toulouse-axel/test/link?NameID=yyy', params=params)
    assert resp.json['err_desc'] == "Data conflict"

    # existing link but person_id is wrong
    link.dui = 'XXX'
    link.person_id = '35'
    link.save()
    with mock_getdata(content, 'RefVerifDui'):
        resp = app.post_json('/toulouse-axel/test/link?NameID=yyy', params=params)
    assert resp.json['err_desc'] == "Data conflict"


@pytest.mark.parametrize('code', [2, 3])
def test_link_endpoint(app, resource, params, code):
    content = '''<PORTAIL>
    <DUI>
        <IDDUI>XXX</IDDUI>
        <IDPERSONNE>42</IDPERSONNE>
        <CODE>%s</CODE>
    </DUI>
</PORTAIL>''' % code
    with mock_getdata(content, 'RefVerifDui'):
        resp = app.post_json('/toulouse-axel/test/link?NameID=yyy', params=params)
    assert set(resp.json.keys()) == set([
        'err', 'link', 'created', 'dui'])
    assert resp.json['err'] == 0
    assert resp.json['dui'] == 'XXX'
    assert resp.json['created'] is True

    # again
    with mock_getdata(content, 'RefVerifDui'):
        resp = app.post_json('/toulouse-axel/test/link?NameID=yyy', params=params)
    assert set(resp.json.keys()) == set([
        'err', 'link', 'created', 'dui'])
    assert resp.json['err'] == 0
    assert resp.json['dui'] == 'XXX'
    assert resp.json['created'] is False  # link already exists


def test_family_info_endpoint_axel_error(app, resource):
    Link.objects.create(resource=resource, name_id='yyy', dui='XXX', person_id='42')
    with mock.patch('passerelle.contrib.toulouse_axel.models.ref_famille_dui') as operation:
        operation.side_effect = AxelError('FooBar')
        resp = app.get('/toulouse-axel/test/family_info?NameID=yyy')
        assert resp.json['err_desc'] == "Axel error: FooBar"


def test_family_info_endpoint_no_result(app, resource):
    resp = app.get('/toulouse-axel/test/family_info?NameID=yyy')
    assert resp.json['err_desc'] == "Person not found"


def test_family_info_endpoint(app, resource):
    Link.objects.create(resource=resource, name_id='yyy', dui='XXX', person_id='42')
    filepath = os.path.join(os.path.dirname(__file__), 'data/toulouse_axel/family_info.xml')
    with open(filepath) as xml:
        content = xml.read()
    with mock_getdata(content, 'RefFamilleDui'):
        resp = app.get('/toulouse-axel/test/family_info?NameID=yyy')
    assert set(resp.json.keys()) == set([
        'err',
        'ADRESSE',
        'CODEMISEAJOUR',
        'DEMATFACTURES',
        'ENFANT',
        'IDDUI',
        'NBENFANTACTIF',
        'NBRLACTIF',
        'REACTUALISATIONENLIGNE',
        'REVENUS',
        'RL1',
        'RL2',
        'SITUATIONFAMILIALE',
        'TELFIXE',
    ])
    assert resp.json['err'] == 0
